#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Facets integration for ZenML.

The Facets integration provides a simple way to visualize post-execution objects
like `PipelineView`, `PipelineRunView` and `StepView`. These objects can be
extended using the `BaseVisualization` class. This integration requires
`facets-overview` be installed in your Python environment.
"""

from zenml.integrations.constants import FACETS
from zenml.integrations.integration import Integration


class FacetsIntegration(Integration):
    """Definition of [Facet](https://pair-code.github.io/facets/) integration for ZenML."""

    NAME = FACETS
    REQUIREMENTS = ["facets-overview>=1.0.0", "IPython"]


FacetsIntegration.check_installation()
