#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Constants for ZenML integrations."""

AIRFLOW = "airflow"
AWS = "aws"
AZURE = "azure"
AZUREML = "azureml"
DASH = "dash"
EVIDENTLY = "evidently"
FACETS = "facets"
FEAST = "feast"
GCP = "gcp"
GCP_SECRETS_MANAGER = "gcp_secrets_manager"
GITHUB = "github"
GRAPHVIZ = "graphviz"
KUBEFLOW = "kubeflow"
LIGHTGBM = "lightgbm"
MLFLOW = "mlflow"
PLOTLY = "plotly"
PYTORCH = "pytorch"
PYTORCH_L = "pytorch_lightning"
S3 = "s3"
SAGEMAKER = "sagemaker"
SCIPY = "scipy"
SELDON = "seldon"
SKLEARN = "sklearn"
SLACK = "slack"
TENSORFLOW = "tensorflow"
VAULT = "vault"
WHYLOGS = "whylogs"
WANDB = "wandb"
VERTEX = "vertex"
NEURAL_PROPHET = "neural_prophet"
HUGGINGFACE = "huggingface"
XGBOOST = "xgboost"
